/* WIMIX Wasser-Impulsmixer Waschmaschinen-Vorschaltgeraet             */
/*                       __    _                                       */                     
/*            __________/ /_  (_)________   ___  __  __                */       
/*           / ___/ ___/ __ \/ / ___/ __ \ / _ \/ / / /                */ 
/*          (__  ) /__/ / / / / /  / /_/ //  __/ /_/ /                 */ 
/*         /____/\___/_/ /_/_/_/  / .___(_)___/\__,_/                  */ 
/*                               /_/                                   */ 
/*                                                                     */ 
/* (c) C.Schirp released under the creative commons BY-NC-SA license   */
/*     for further details see                                         */ 
/*     http://creativecommons.org/licenses/by-nc-sa/3.0/               */ 
/*     schirp.eu must be named in derived software coyright notes      */
/*                                                                     */
/*    V1.0  Erstausgabe 08.02.2011                                     */
/*    V1.1  Zahlenformat gendert, AD-Werte skaliert                   */
/*    V1.11 Update KW-Temperatur auch im PURGE-Status, Shutdown        */
/*          mit mehr Stromsparen                                       */
/*    V1.2  Zulaufmengenbegrenzung hinzu, Fehlertexte                  */
/*                                                                     */

#include <c:\projekte\picsel\cc5\16F819.h>
#include <c:\projekte\picsel\cc5\int16CXX.h>
#include <c:\projekte\picsel\cc5\inline.h>

#pragma config = 0b.0010.1111.0011.1000 



/* ====================== Port- Defines ====================== */

#define LCD_RS    0b.00001000             //RS-Pin ist Bit3, LCD Register Select
#define LCD_CS    0b.00000100             //E-Pin ist Bit2, LCD Chip Select
#pragma char PORTQ @ PORTB                //PORT auswhlen, DB7-Bit7, DB6-Bit6, DB5-Bit5, DB4-Bit4

#define KW_VENTIL PORTB.0			// Kaltwasser-Ventilausgang
#define WW_VENTIL PORTB.1			// Warmwasser-Ventilausgang
#define VENTIL_IN1 PORTA.1			// Ventileinang 1 
#define TASTE_PROG PORTA.3			// Tasteneingang Programm
#define TASTE_VORW PORTA.4			// Tasteneingang Vorwaesche
#define LCD_BACKLIT PORTA.6		// Backlight
#define LED_OFF PORTA.7			// LED


/* ====================== Kommandos fr das LCD-Display ====================== */
#define CLEAR_DISPLAY 0x01
#define OFF_DISPLAY   0x08
#define CURSOR_OFF    0x0C
#define CURSOR_ON     0x0E
#define CURSOR_BLINK  0x0F
#define SHIFT_LEFT    0x18
#define SHIFT_RIGHT   0x1C

/* ====================== Programmkonstanten ====================== */
#define LCD_LENGTH    17
#define Hund Text+0
#define Tens Text+1
#define Ones Text+2

#define CGRAM_LEN	24  //Anzahl zu bertragender Zeichen

#define ST_IDLE		0
#define ST_MIX		1
#define ST_PURGE	2
#define ST_FAIL		3

#define TEMP_OFFS      237	// Offset AD-Wert bei 0Grad
#define TEMP_MAX       121	// Maximaler AD-Wert
#define TEMP_TM0       20 //1260	// Trommel-Integral Startwert (Temperatur x Masse)  	
#define TEMP_CNT0	1 //30	// Vorbelegung der Trommel fr TM0

#define TEMPSCALE 175	// Steigung ADC, KTY81 und Rvor: 0,64/Bit -> 0,63*256 = 161 (gemessen: 177)

#define MIXZEIT_MAX 900 	// Zeit fr Warmwasser-Zulauf 15 Minuten = 15x 60 Sekunden
#define VORWZEIT_PAUSE 300	// Zeit ohne Wasserbezug, ab der auf Haupwaesche geschaltet wird
#define VORW_TEMP 20		// Zieltemperatur fr Vorwsche
#define ZULAUFZEIT_MAX 	860	// Maximale Zeit mit aktivem Wasserzulauf... Wasserstop-Funktion (100 l bei 7l/min)

#define LCD_LIGHT	15	// Zeit fr Backlight an in Sekunden

#define F_NONE	0	// kein Fehler
#define F_LOWT 	1	// untertemperatur
#define F_OVERT	2	// Uebertemp
#define F_FUELLZEIT 3	// Fuellzeit lnger als ZULAUFZEIT_MAX

#define VENTIL_CNT_MAX 6 // Anzahl 2ms-Perioden fr Ventileingangserkennung

#define VENTIL_AUF 0	// defines fr die Low-Aktiven Ventile
#define VENTIL_ZU 1

#define VENTIL_OVERLAP	50 // Anzahl 2,5ms-Scheiben, die beide Ventile beim Wechsel offen sind


/* ====================== Funktionsprototypen, Makros etc. ====================== */
#define LCDCS_ON() { NOP;  NOP; lcdport|=LCD_CS; PORTQ=lcdport; }
#define LCDCS_OFF() { NOP;  NOP; lcdport&=~LCD_CS; PORTQ=lcdport; }
#define LCDRS_ON() { NOP;  NOP; lcdport|=LCD_RS; PORTQ=lcdport; }
#define LCDRS_OFF() { NOP; NOP; lcdport&=~LCD_RS; PORTQ=lcdport; }

#define Wait() Delay1ms(10)

/* ======================  Variablen ====================== */

bank0 uns16 ZulaufTemp;	// aktuell gemessene ZulaufTemperatur
bank0 uns8 KWTemp;	// zuletzt gemessene Kaltwasser Temperatur
bank0 uns8 WWTemp;	// zuletzt gemessene Warmwasser Temperatur
bank0 uns8 TrommelTemp;// berechnete TrommelTemperatur
bank0 uns16 TrommelInt;	// Trommeltemperatur-Integral
bank0 uns16 TrommelCount;	// Trommeltemperatur-Integral
bank0 uns8 ZielTemp;	// Zieltemperatur (=Programm)

int16 med_a;
int16 med_b;
int16 med_c;

bank0 uns16 Gesamtzeit;	// gemessene Zeit seit Start in Sekunden
bank0 uns16 SubSeczeit;	// Zeitslots innerhalb einer Sekunde
bank0 uns16 Wasserloszeit;// Zeit ohne Wasserbezug
bank0 uns16 Zulaufzeit;	// Zeit mit aktivem Wasserbezug

uns8 akt_status;	// Zentrale Statusvariable
uns8 fcode;		// Fehlercode
uns8 v1_count;		// Hilfsvariable zum Entprellen der 50Hz-Ventilsignale
//uns8 v2_count;

uns8 v_overl_cnt;	// Rckwrtszhler fr Ventil-berlappzeit
bit  v_overl_req;	// welches Ventil soll verzgert getoggelt werden? 0=KW 1=WW
bit overl_ak_kw;	// Aknowledge-Bits fr Overlap kalt->warm 
bit overl_ak_wk;	// warm->kalt


bit V_IN1_old;
//bit V_IN2_old;
bit V_IN1;		// Entprellter Ventil 1 Status
//bit V_IN2;		// Entprellter Ventil 2 Status
bit T_PROG;		// Entprellte Taste Prog
bit T_VORW;		// Entprellte Taste Vorwsche

bit VORWAESCHE;		// Vorwaesche gewuenscht
bit Vorw_Pause_Merker;	// Zulaufpause Vorwsche erreicht		
bit T_PROG_old;
bit T_VORW_old;

bit changed;		// Taste oder Ventileingang hat sich gendert
bit Update_Prog;	// Update-Request fr LCD-Programmzeile


unsigned char Text[LCD_LENGTH];	// Speicher fr eine LCD-Zeile incl Nullterminierung
uns8 lcdport;		// temp-variable der LCD-Routinen
uns8 lcdlight;		// Backlight-Counter

uns8 q[4];	// Ergebnisvariable fr divide


const unsigned char chardefs[] = {
//Adresse 0 im Character RAM
// 'raute'
0, 0x04, 0x0A, 0x11, 0x0A, 0x04, 0, 0,

//Adresse 1 im Character RAM
// 'quadrat'
0, 0, 0x0e, 0x0a, 0x0e, 0, 0, 0,

//Adresse 2 im Character RAM
// 'gefuelltes quadrat'
0, 0, 0x0e, 0x0e, 0x0e, 0, 0, 0,

};


/*------------------------------------------------------------------*/
/*   II  N   N TTTTT                                                */
/*   II  NN  N   T                                                  */
/*   II  N N N   T                                                  */
/*   II  N  NN   T                                                  */
/*   II  N   N   T                                                  */
/*------------------------------------------------------------------*/
#pragma origin 4
interrupt int_server(void)
{
	int_save_registers
	
	// Interrupt-Code
	
	int_restore_registers
}



/*------------------------------------------------------------------*/
/*   II  N   N  II TTTTT                                            */
/*   II  NN  N  II   T                                              */
/*   II  N N N  II   T                                              */
/*   II  N  NN  II   T                                              */
/*   II  N   N  II   T                                              */
/*------------------------------------------------------------------*/

void init (void)
{
	/* Oszillatorblock */
	
	OSCCON = 0b0110.0000;	// 4MHz Clock - wie 16F628 mit INTRC
	
	/* PORT A Config */
	PORTA =  0x00;	
	TRISA =  0b0011.1111;	// RA0-5 input , RA6,7 Output
	
	/* Port B Config */
	PORTB =  0b0000.0011;	// RB0,1 sind die Ventilausgnge... low aktiv
	TRISB =  0b0000.0000;	// alles Output

	/* ADC */
	ADCON0=	 0b1000.0001;	// Fosc...  Channel AN0, Stopped, AD_ON
	ADCON1 = 0b1100.1110;	// Rechtsbuendig, AD_clock/2, DDDA-Konfig

	/*sonstiges*/
	T1CON = 0x00;		// Timer disabled
	OPTION_REG = 0x84;	/* pull-up porta, falling edge int, FOSC/4 source, no WDT prescaler, prescale=32 */
	INTCON = 0b.0000.0000; 	// int disabled und gelscht 
	
	LCD_BACKLIT = 1;	// erstmal dunkel machen

	/* Variablen initialisieren */

	Text[0]= 0;
	akt_status = ST_MIX;	
	TrommelInt = TEMP_TM0;
	TrommelCount = TEMP_CNT0;
	TrommelTemp = 20;
	
	ZielTemp = 30;	// Default-Temperatur 30 Grad
	VORWAESCHE = 0;
	
	Vorw_Pause_Merker =0;
	
	V_IN1_old = 0;	// Eingangsmerker aktualisieren
	//V_IN2_old = 0;
	V_IN1 = 0;
	//V_IN2 = 0;	
	T_PROG = 0;		
	T_VORW = 0;
	T_PROG_old = TASTE_PROG;
	T_VORW_old = TASTE_VORW;
	
	
	overl_ak_kw = 0; // Aknowledge-Bits lschen
	overl_ak_wk = 0;

	v1_count = 0;		
	//v2_count = 0;
	
	v_overl_cnt = 0;
	
	Update_Prog = 1;	// Zeile eins auch updaten
	
	Gesamtzeit = 0;
	SubSeczeit = 0;
	
	Zulaufzeit = 0;	        // Aufsummierung der Zulaufzeit zuruecksetzen
	
	lcdlight = LCD_LIGHT;	// Licht anmachen
}


/*------------------------------------------------------------------*/
/*   LL    CCCC DDD                                                 */
/*   LL   CC    DD DD                                               */
/*   LL   CC    DD  DD                                              */
/*   LL   CC    DD DD                                               */
/*   LLLLL CCCC DDD                                                 */
/*------------------------------------------------------------------*/

void Delay1ms(uns16 ms)
{

	while(ms)				// Schleife verlassen wen ms=0 ist
	{
	    OPTION = 2; 		// Vorteiler auf 8 einstellen
	    TMR0 = 131; 		// 125 * 8 = 1000 (= 1 ms)
	    while (TMR0);		// abwarten einer Milisekunde
	    ms--;				// "ms" mit jeder Milisekunde ernidrigen
	}

}

//*******************************************
// Schreibt ein Datenbyte in die LCD-Anzeige
// Schreibt das Byte in zwei 4 Bit Nibbles
//*******************************************

void LCDWriteByte (unsigned char data)
{
#pragma updateBank 0 /* OFF, all cores */
 unsigned char buff;
 lcdport= PORTQ; 		//alten Wert merken
 
 LCDCS_OFF();
 buff=data & 0xF0;		//High Nibble des Bytes, Rest lschen
 lcdport=lcdport&0x0F;		//Hier D4-D7 lschen zum
 lcdport|=buff;			//reinodern vom High Nibble
 lcdport|=LCD_CS;
 LCDCS_ON();	      		//High Nibble schreiben

 LCDCS_OFF();
 buff=data<<4;        	//Low Nibble des Bytes hochschieben
 buff &= 0xF0;		// nachgeschobenes Loeschen
 lcdport=lcdport & 0x0F;	//D4-D7  lschen zum
 lcdport|=buff;			//reinodern vom Low Nibble
 LCDCS_ON();	      		//Low Nibble schreiben
 LCDCS_OFF();
 Delay1ms(1);       		// warten
#pragma updateBank 1 /* ON, all cores */
}

//*****************************************
// Commando ans Display senden
//*****************************************
void LCDCommand(unsigned char mode)
{
#pragma updateBank 0 /* OFF, all cores */
 lcdport= PORTQ; 		//alten Wert merken
 LCDRS_OFF(); 			//Controlregister
 LCDWriteByte(mode);
 LCDRS_ON(); 			//Zurck auf Datenregister
#pragma updateBank 1 /* ON, all cores */
}

//*****************************************
// Lscht die LCD Anzeige
// Cursor steht dann auf Position 1,Zeile 1
// Gilt auch fr mehrzeilige Anzeigen
//*****************************************
void LCDCls(void)
{
 lcdport= PORTQ; 		//alten Wert merken
 LCDRS_OFF(); //Controlregister
 LCDWriteByte(0x01); //CLS
 LCDRS_ON(); //Umschalten auf Datenregister
 Wait();
}
//**************************************************
//Schreibt einen Text in die LCD-Anzeige
//**************************************************
void LCDString(const char *__putstr)
{
//#pragma updateBank 0 /* OFF, all cores */
 char __ps;

 __ps = *__putstr;

  while(__ps>0)
   {
    __putstr++;
    if (__ps== 0) break;
   LCDWriteByte(__ps);
    __ps = *__putstr;
   }
//#pragma updateBank 1 /* ON, all cores */
}
//****************************************************
// Springt auf Zeile x und Position y der LCD-Anzeige
// Gltig fuer 2 Zeilen : 0,1.
//****************************************************
void LCDPos(unsigned char zeile,unsigned char position)
{
#pragma updateBank 0 /* OFF, all cores */
	 unsigned char buff;
	
	// 0x80 wegen Display Adress Set.
	// Erste Position ist bei mir  0.
	
	if (zeile)
	{
		buff=0xC0+position;
	}
	else
	{
		buff=0x80+position;
	}
	
	 LCDCommand(buff);
#pragma updateBank 1 /* ON, all cores */
}


//*********************************************
//Stellt die Funktionsweise der LCD Anzeige ein
//*********************************************

void LCDInit(void)
{
#pragma updateBank 0 /* OFF, all cores */
 int i;
 
 lcdport= PORTQ; 	//alten Wert merken

 LCDRS_OFF();  //Control Register
 LCDCS_OFF();  //LCD-Anzeige abwhlen

//Software Reset laut Hitachi Datenblatt
 Delay1ms(60); //Mglichen Power On Reset abwarten
 for(i=0; i<3; i++)
  {
   lcdport=lcdport&0x0F;//D4-D7 lschen
   lcdport|=0x30;	//Wert reinodern
   PORTQ=lcdport;	//Auf den Port legen
   LCDCS_OFF(); //LCD-Anzeige abwhlen
   NOP;
   LCDCS_ON();	//High Nibble schreiben
   Delay1ms(5); //Ein bichen warten auf die Anzeige
  } 
//Ende Software Reset


//Der erste Schreibzugriff nach dem Reset ist ein 8 Bit
//Zugriff!! Also nur einmal schreiben, nicht Low-High Nibble.

//Um die 4Bit Operation einzuschalten
 lcdport=0x23;	//4 Bit Operation 2 Zeilen, untere Bits an wg. Ventilausgngen	

 PORTQ=lcdport;	//Auf den Port legen - ACHTUNG, Zerstrt auch die Bits0..3!
 NOP;

 LCDCS_ON();	//High Nibble schreiben
 NOP;
 NOP; 
 LCDCS_OFF();
 Wait();
 LCDWriteByte(0x2C);    //4 Bit Operation,2 Line,5x7 Font  
 Wait();
 LCDWriteByte(0x14); 	//Cursor Move,Right Shift
 Wait();

 LCDWriteByte(0x0C); 	//Display on, Cursor off ,Cursor no blink
 Wait();

 LCDWriteByte(0x06); 	//Increment, Display Freeze
 Wait();
 LCDWriteByte(0x02); 	//Cursor Home
 Wait();
 LCDRS_ON();		//LCD-Datenregister

 LCDCommand(CLEAR_DISPLAY);
 Wait();
 
//Die eigenen Character ins CGRAM schreiben

//Der interne Adresszhler des LCD-Displays luft
//nach jedem Schreibzugriff automatisch hoch

 LCDCommand(0x48);//CG_RAM Adresse(1) setzen

 for(i=0; i<CGRAM_LEN; i++) //8 Codes pro Zeichen bertragen
  {
   LCDWriteByte(chardefs[i]);
  }


 LCDCommand(0x80); //DD_RAM Adresse 0 setzen
 
#pragma updateBank 1 /* ON, all cores */
}

#ifdef _DEBUG
/*------------------------------------------------------------------*/
/*  Konvertiert 16 Bit Binaer zu Hexadezimal-ASCII 4-stellig        */
/*------------------------------------------------------------------*/
void Dez4Hex(uns16 zahl)
{
	char  count, help;

	for (count = 0; count < 4; count++)
	{
		
		help = zahl & 0x000F;
		
		if (help > 9)
		{
			Text[3 - count] = help + 0x37;
		}
		else
		{
			Text[3 - count] = help + 0x30;
		}
		zahl >>=4;
	}
}
#endif

/*------------------------------------------------------------------*/
/*  Konvertiert 16 Bit Binaer zu Dezimal-ASCII 3-stellig            */
/*------------------------------------------------------------------*/
void bin2dec999(uns16 Num)
{
#asm
;Binary to decimal conversion (0..999)
;
;Input: Num+1:Num
;Output Hund:Tens:Ones (-> per define auf Text[])
;
;If Input > 999 Output will roll over, e.g.
;for input=5678 output=678.
;
;
;Size: 34 instructions
;Execution  time  (max) including return:
;22+5*9-1+5*6-1+4*3-1+2 = 108 cycles
;
;5-July-2000 by Nikolai Golovchenko


        movf Num+1, w
        addlw 241
        addwf Num+1, w
        movwf Hund      ;b_2 = 2a_2 - 15

        addwf Hund, w
        addwf Hund, w
        addlw 253
        movwf Tens
        swapf Num, w
        andlw 0x0F
        addwf Tens, f
        addwf Tens, f   ;b_1 = 6a_2 + 2a_1 - 48

        addwf Num+1, w
        sublw 251
        movwf Ones
        addwf Ones, f
        addwf Ones, f
        addwf Ones, f
        movf Num, w
        andlw 0x0F
        addwf Ones, f   ;b_0 = a_0 - 4(a_2 + a_1) - 20

        movlw 10
bin2dec999a                     ;9 cycles max
        addwf Ones, f
        decf Tens, f
        btfss 3,Carry
         goto bin2dec999a

bin2dec999b                     ;6 cycles max
        addwf Tens, f
        decf Hund, f
        btfss 3,Carry
         goto bin2dec999b

bin2dec999c                     ;3 cycles max
        addwf Hund, f
        btfss 3,Carry
         goto bin2dec999c

#endasm
	Text[0] +=0x30;
	Text[1] +=0x30;
	Text[2] +=0x30;
	Text[3] =0;
	
}


/*------------------------------------------------------------------*/
/*                                                                  */
/*                                                                  */
/*                                                                  */
/*                                                                  */
/*                                                                  */
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/
/*   MM     MM    AA  TTTTTT HH   HH                                */
/*   MMM   MMM  AA  AA  TT   HH   HH                                */
/*   MM M M MM  AAAAAA  TT   HHHHHHH                                */
/*   MM  M  MM  AA  AA  TT   HH   HH                                */
/*   Mm     MM  AA  AA  TT   HH   HH                                */
/*------------------------------------------------------------------*/



void mul16(uns16 n1, uns16 n2)
{

#asm
;*******************************************************************
; Double Precision Multiplication
;
; n_2 : n_1 * n_4 : n_3 -> q_4:q_3:q_2:q_1
;
; Standard shift and add.
; Execution time: 215 to 295 clock cycles.
; Code space: 22 locations
;
; Cleaned up and corrected version from Microchip Ap note by BF.
; Note: Ap note has errors! Additional mods by Scott Dattalo.
;
;*******************************************************************

	clrf q+3
	clrf q+2
	clrf q+1
	clrf q	
	bsf q+1, 7	
m1:
	rrf n1+1, f
	rrf n1, f
	btfss 3,Carry  ; skpc
	goto m2
	movf n2, w
	addwf q+2, f
	movf n2+1, w
	btfsc 3,Carry  ; skpnc
	incfsz n2+1, w
	addwf q+3, f
m2:
	rrf q+3, f
	rrf q+2, f
	rrf q+1, f
	rrf q, f
	btfss   3,Carry   ;skpc
	goto m1
#endasm

}
        
        
void divide()
{       
        
	uns16 REM;  	// Divisionsrest -> Discard
	uns8 LOOPCOUNT;

#asm    
        
;Inputs:
;   Dividend - AARGB0:AARGB1:AARGB2 (0 - most significant!)
;   Divisor  - BARGB0:BARGB1
;Temporary:
;   Counter  - LOOPCOUNT
;   Remainder- REMB0:REMB1
;Output:
;   Quotient - AARGB0:AARGB1:AARGB2	  
        
FXD2416U:
            clrf REM
            clrf REM+1
            movlw .24
            movwf LOOPCOUNT
LOOPU2416:
            rlf q, W           ;shift dividend left to move next bit to remainder
            rlf q+1, F           ;
            rlf q+2, F           ;
        
            rlf REM+1, F            ;shift carry (next dividend bit) into remainder
            rlf REM, F
        
            rlf q, F           ;finish shifting the dividend and save  carry in AARGB2.0,
                                    ;since remainder can be 17 bit long in some cases
                                    ;(e.g. 0x800000/0xFFFF). This bit will also serve
                                    ;as the next result bit.
             
            movf TrommelCount, W          ;substract divisor from 16-bit remainder
            subwf REM+1, F          ;
            movf TrommelCount+1, W          ;
            btfss STATUS, Carry         ;
            incfsz TrommelCount+1, W        ;
            subwf REM, F          ;

    ;here we also need to take into account the 17th bit of remainder, which
    ;is in AARGB2.0. If we don't have a borrow after subtracting from lower
    ;16 bits of remainder, then there is no borrow regardless of 17th bit 
    ;value. But, if we have the borrow, then that will depend on 17th bit 
    ;value. If it is 1, then no final borrow will occur. If it is 0, borrow
    ;will occur. These values match the borrow flag polarity.

            btfsc 3,Carry           ;skpnc - if no borrow after 16 bit subtraction
             bsf q, 0          ;then there is no borrow in result. Overwrite
                                    ;AARGB2.0 with 1 to indicate no
                                    ;borrow.
                                    ;if borrow did occur, AARGB2.0 already
                                    ;holds the final borrow value (0-borrow,
                                    ;1-no borrow)

            btfsc q, 0         ;if no borrow after 17-bit subtraction
             goto UOK46LL           ;skip remainder restoration.

            addwf REM, F          ;restore higher byte of remainder. (w 
                                    ;contains the value subtracted from it
                                    ;previously)
            movf TrommelCount, W          ;restore lower byte of remainder
            addwf REM+1, F          ;

UOK46LL:

            decfsz LOOPCOUNT, f     ;decrement counter
             goto LOOPU2416         ;and repeat the loop if not zero.

 #endasm         
            
}

void MedianFilter(void)
{
#pragma updateBank 0 /* OFF, all cores */

/* Beginn Median-Filter */

  if (med_a>med_b)
   if (med_a<med_c)
    ZulaufTemp = med_a;
   else
    if (med_b<med_c)
     ZulaufTemp = med_c;
    else
     ZulaufTemp = med_b;
  else
   if (med_a>med_c)
    ZulaufTemp = med_a;
   else
    if (med_b<med_c)
     ZulaufTemp = med_b;
    else
     ZulaufTemp = med_c;
     
#pragma updateBank 1 /* OFF, all cores */
}



/*------------------------------------------------------------------*/
/*     AA    PPPPP  PPPPP                                           */
/*   AA  AA  PP  PP PP  PP                                          */
/*  AAAAAAA  PPPPP  PPPPP                                           */
/*  AA   AA  PP     PP                                              */
/*  AA   AA  PP     PP                                              */
/*------------------------------------------------------------------*/


//
// ZulaufTempmessung ber ADC-Eingang AN0
// AD-Wert wird geprueft, offsetkorrigiert, skaliert in "ZulaufTemp"  geschrieben
//
// liest: -
// schreibt: ZulaufTemp, akt_status, fcode
//
void temp_messen(void)
{
//#pragma updateBank 0 /* ON, all cores */

uns16 
		
	ADCON0=	 0b1000.0001;	// AD-Wandler rechtsbuendig, VDD=ref,  Channel AN0, Stopped, AD_ON
	
	// Anmerkung zur Geschwindigkeit: AD Starten, Ergebnis kopieren, dann warten bringt nichts, da
	// nach jeder AD-Wandlung 16 Zyklen vor Neustart des ADC gewartet werden muss.

	GO = 1;			// Wandlung starten
	while (GO);		// Auf ADC warten
	
	// Overflow abfangen
	med_a.high8 = ADRESH;
	med_a.low8 = ADRESL;
	
	GO = 1;			// Wandlung starten
	while (GO);		// Auf ADC warten
	
	// Overflow abfangen
	med_b.high8 = ADRESH;
	med_b.low8 = ADRESL;
	
	GO = 1;			// Wandlung starten
	while (GO);		// Auf ADC warten
	
	// Overflow abfangen
	med_c.high8 = ADRESH;
	med_c.low8 = ADRESL;
	
	MedianFilter(); // AD-Werte Filtern
	
	// Wertebereich prfen, Offset runter
	if ( ZulaufTemp > TEMP_OFFS )
	{
		// Okay, offset abrechnen
		ZulaufTemp -= TEMP_OFFS;
		
		// Obergrenze testen
		if ( ZulaufTemp > TEMP_MAX )
		{
			// ZulaufTemp zu gross
			akt_status = ST_FAIL;
			fcode = F_OVERT; 	
		}
	}
	else // ZulaufTemp zu klein
	{
		akt_status = ST_FAIL;
		fcode = F_LOWT;
	}
	
	// --- skalieren ---
	// Wir rechnen Fixedpoint mit 8 Nachkommabits
	
	mul16(ZulaufTemp,TEMPSCALE);	// Skalierungsfaktor einrechnen: 1 Bit -> 0,64
	
	// Runden: wenn oberstes Bit des Nachkommateils gesetzt ist, Vorkommateil inkrementieren
	if (q[0].7)
	{
		q[1]++;
	}
	
	ZulaufTemp.low8  = q[1];	// Die oberen 8 Bit sind jetzt der Vorkomma-Anteil
	ZulaufTemp.high8 = 0;

	
	
//#pragma updateBank 1 /* ON, all cores */		
}

//
// --- Eingnge lesen ---
// auf Tasten oder Ventileingnge testen und Entprellen, 2ms warten-> Mastertakt!
//
// liest: 
// schreibt: V_IN1, T_VORW, T_PROG
//

void check_input(void)
{
#pragma updateBank 0 /* OFF, all cores */
	
	changed = 0;
	
	// ------ Ventileingnge --------------
	
	// Ventileingnge am uC sind Low-Aktiv!
	if (!VENTIL_IN1)
	{
		if (v1_count < VENTIL_CNT_MAX)
		{
			v1_count += 2;
		}
	}
	else if (v1_count >= 1)
	{
		v1_count--;	
	}
	
	if (v1_count >= VENTIL_CNT_MAX)
	{
		// Eingang ist aktiv!
		V_IN1 = 1;
	}
	else if (v1_count == 0)
	{
		// Eingang ist nicht aktiv
		V_IN1 = 0;
	}
	
	if (V_IN1_old ^ V_IN1)
	{
				
		changed = 1;
		V_IN1_old = V_IN1;
	}
	//--- 
	
	
	// Ventileingnge am uC sind Low-Aktiv!
	/*
	if (!VENTIL_IN2)
	{
		if (v2_count < VENTIL_CNT_MAX)
		{
			v2_count += 2;
		}
	}
	else if (v2_count >= 1)
	{
		v2_count--;	
	}
	
	if (v2_count >= VENTIL_CNT_MAX)
	{
		// Eingang ist aktiv!
		V_IN2 = 1;
	}
	else if (v2_count == 0)
	{
		// Eingang ist nicht aktiv
		V_IN2 = 0;
	}
	
	if (V_IN2_old ^ V_IN2)
	{
				
		changed = 1;
		V_IN2_old = V_IN2;
	}
	
	*/
	// ------ Tasteneingnge --------------
	if (T_PROG_old ^ TASTE_PROG)
	{
		changed = 1;
		T_PROG = !TASTE_PROG;
		T_PROG_old = TASTE_PROG;
	}
	
	if (T_VORW_old ^ TASTE_VORW)
	{
		changed = 1;
		T_VORW = !TASTE_VORW;
		T_VORW_old = TASTE_VORW;
	}
	
#pragma updateBank 1 /* ON, all cores */		
}

//
// --- Tasten auswerten  ---
//
// Auswertung entprellter Tasten, Zielprogramm und Vorwsche festlegen
// setzt Entrpellte Tasten auf 0 zurck als Besttigung
//
// liest: T_PROG, T_VORW
// schreibt: ZielTemp, VORWAESCHE, T_PROG, T_VORW 
// 
void update_tasten(void)
{
#pragma updateBank 0 /* OFF, all cores */

	// Tastenbedienung immer mglich
	if (T_PROG)
	{
		if (ZielTemp == 0)
		{
			ZielTemp = 30;
		}
		else if (ZielTemp < 60)
		{
			ZielTemp += 10;
		}
		else if (ZielTemp == 60)
		{
			ZielTemp = 0;
		}
		
		T_PROG = 0; // Taste besttigen
		Update_Prog = 1;
		lcdlight = LCD_LIGHT;	// Licht anmachen
	}
	
	if (T_VORW)
	{
		VORWAESCHE ^= 1;
		T_VORW = 0; // Taste besttigen
		Update_Prog = 1;
		lcdlight = LCD_LIGHT;	// Licht anmachen
	}
		
#pragma updateBank 1 /* ON, all cores */		
}


//
// --- Statemachine bedienen ---
//
// Zeitrechnung, Status bzw. Programmfortschritt, Trommeltemperaturberechnung
//
// liest: SubSeczeit, Gesamtzeit, VORWAESCHE, TrommelInt, TrommelCount, changed, akt_status
// schreibt: SubSeczeit, Gesamtzeit, changed, VORWAESCHE, TrommelInt, TrommelCount, akt_status, TrommelTemp, q[], Zulaufzeit
//
void update_state(void)
{
#pragma updateBank 0 /* OFF, all cores */

uns8 temp2;

	// --- Zeitrechnung updaten ---

	SubSeczeit++;
	
	// 1 Sec um? (gemessener Wert: 318 ticks / 3,1ms Ticktime)
	if (SubSeczeit > 318)
	{
		Gesamtzeit++;	// Gesamtzeit updaten
		SubSeczeit = 0;
		changed = 1;	// mal nen update machen
		
		// Zeit mit/ohne Wasserbezug mitstoppen
		if (V_IN1)
		{
			Wasserloszeit=0;
			Zulaufzeit++;
		}
		else
		{
			Wasserloszeit++;
		}
	}
	
	
	// --- Hier kommt die Statemachine ---
	
	
	if (VORWAESCHE)
	{
		if (Wasserloszeit >= VORWZEIT_PAUSE)
		{
			// Wenn die Pausenzeit fr Wassereinlauf bei der Vorwsche erreicht ist,
			// beim nchsten Zulauf die Hauptwaschgangzeit starten, Merker setzen
			
			Vorw_Pause_Merker = 1;
		}
		
		// Wenn die Zulaufpause erreicht war 
		if (Vorw_Pause_Merker)
		{
			// ...und wieder Wasser angefordert wird, ist es der Hauptwaschgang
			if (V_IN1)
			{
				// Wenn die Vorwsche um ist, einfach eine neue Zeitrechnung 
				// ohne Vorwsche starten -> Hauptwaschgang
				Gesamtzeit = 0;
				VORWAESCHE = 0;
				SubSeczeit = 0;
				TrommelInt = TEMP_TM0;
				TrommelCount = TEMP_CNT0;
				changed = 1;
			}
		}
	}
	
	if (Gesamtzeit < MIXZEIT_MAX)
	{
		// wenn kein Fehler, Status = mixen
		if (akt_status != ST_FAIL)
		{
			akt_status = ST_MIX;
		}
		
		// --- aktuelle Temperatur berechnen ---
	
		//Nur bei Umspringen von Sekunde oder Eingngen (Todo: beim Tastendrcken erfolgt Integralupdate...)
		if (changed)
		{
			if (Gesamtzeit.1)
			{
				// Wenn der Zulauf an ist, ist auch eines unserer Ventile auf
				if (V_IN1)
				{
					
					// Kaltwasser-Temperatur korrigieren
					if (KW_VENTIL == VENTIL_AUF)
					{
						// wenn die Zulauftemperatur KW gestiegen ist (nach Umschalten von Warm z.B.)
						// den alten KW-Wert nehmen
						if (ZulaufTemp.low8 > KWTemp)
						{
							ZulaufTemp.low8  =  KWTemp;
						}
						else
						{
							KWTemp = ZulaufTemp.low8;
						}
					}
					
					// Warmwasser-Temperatur korrigieren
					if (WW_VENTIL == VENTIL_AUF)
					{
						// wenn die Zulauftemperatur WW gefallen ist (nach Umschalten von Kalt z.B.)
						// den alten WW-Wert nehmen
						if (ZulaufTemp.low8 < WWTemp)
						{
							ZulaufTemp.low8 = WWTemp;
						}
						else // Temperatur gleich oder gestiegen
						{
							//Anstieg Beschleunigen wg. Trgheit der Messung
							temp2 = ZulaufTemp.low8 - WWTemp;
							temp2 *=2; // Differenz x 2 zur Beschleunigung
							WWTemp = ZulaufTemp.low8 + temp2;
							
							TrommelInt += temp2;
							
						}
					}
					
					
					TrommelInt += ZulaufTemp;
					TrommelCount++;
					
					q[0]=TrommelInt.low8;
					q[1]=TrommelInt.high8;
					q[2] =0;
					
					divide();
					
					TrommelTemp.high8=q[1];
					TrommelTemp.low8=q[0];
					
				}	
			}
		}
		
	}
	else 
	{
		// wenn kein Fehler, Status = mixen
		if (akt_status != ST_FAIL)
		{
			akt_status =ST_PURGE;
		}
		
	}
	
	if (Zulaufzeit > ZULAUFZEIT_MAX)
	{
		akt_status = ST_FAIL;	// Fehlerstatus setzen
		fcode = F_FUELLZEIT; 
	}
		

#pragma updateBank 1 /* ON, all cores */
}

//
// --- Ventilstellung updaten ---
//
// Ansteuerung der Ventile, zweipunkt-Regler
//
// liest: V_IN1, ZielTemp, akt_status, Zulaufzeit,
// schreibt: KW_VENTIL, WW_VENTIL
//
void update_vent(void)
{
#pragma updateBank 0 /* OFF, all cores */

bit temp;	// Hilfsvariable
bit temp2;	// Hilfsvariable2
uns16 Solltemp;


	// Abfangjaeger Wasserstop: bei berschreiten der ZULAUFZEIT_MAX alle Ventile zu!
	if (fcode == F_FUELLZEIT)
	{
		temp = 0;	// Fehler, also Ventile schliessen
	}
	else // ansonsten abhngig vom Ventilstatus -> Standardvorgehen
	{
		temp = V_IN1;
	}

	
	// Einlaufventil auf? Dann regeln
	if (temp)
	{
		
		// steht noch ein berlapp-Vorgang an?
		if (v_overl_cnt > 0)
		{
			v_overl_cnt --; // Zhler dekrementieren
			
			// jetzt null? Dann gefordertes Ventil schliessen
			if (v_overl_cnt == 0)
			{
				if (v_overl_req)
				{
					WW_VENTIL = VENTIL_ZU;	
				}
				else
				{
					KW_VENTIL = VENTIL_ZU;
				}
			}
			
		}
		
		// Status Purge oder Fail zusammenfassen
		temp = (akt_status == ST_PURGE);
		temp2 =(akt_status == ST_FAIL);
		temp2 |= temp;
		
		// einfachster Fall: Kalt-Programm 
		if (ZielTemp == 0)
		{
			KW_VENTIL = VENTIL_AUF;
			overl_ak_kw = 0;	// overlap kalt->warm freigeben	
			NOP;			// zwei Nops wegen Timing read-mod-write am port
			NOP;
			WW_VENTIL = VENTIL_ZU;
		}
		else if (akt_status == ST_MIX)  // ansonsten: Regeln  in abh. vom State
		{
			
			// bei Vorwsche usprngliche Temperatur merken, Vorwaschtemp einstellen
			if (VORWAESCHE)
			{
				Solltemp = ZielTemp;
				ZielTemp = VORW_TEMP;	
			}
			
			if (TrommelTemp < ZielTemp) // Unter Ziel? Warm auf
			{
				WW_VENTIL = VENTIL_AUF;
				//NOP;				// zwei Nops wegen Timing read-mod-write am port
				//NOP;
				//KW_VENTIL = VENTIL_ZU;
				
				// wenn der overlap noch nicht besttigt ist
				if (!overl_ak_kw) 
				{
					v_overl_cnt = VENTIL_OVERLAP;
					v_overl_req = 0;
					overl_ak_kw = 1;	// kalt->warm aktiv
					overl_ak_wk = 0;	// warm->kalt freigeben
				}	
			}
			else if (TrommelTemp > ZielTemp) // ber Ziel: Kalt auf
			{
				KW_VENTIL = VENTIL_AUF;
				//NOP;				// zwei Nops wegen Timing read-mod-write am port
				//NOP;
				//WW_VENTIL = VENTIL_ZU;

				// wenn der overlap noch nicht besttigt ist
				if (!overl_ak_wk) 
				{
					v_overl_cnt = VENTIL_OVERLAP;
					v_overl_req = 1;
					overl_ak_wk = 1;	// warm->kalt aktiv
					overl_ak_kw = 0;	// kalt->warm freigeben
				}	
			}
			else  
			{
				// Zieltemp genau getroffen... wenn kein Ventil auf ist-> warmwasser. Kommt vor, wenn Zulauf
				// zu war und die Temperatur zufllig genau auf der Mitte stand.
				// Im Normalbetrieb (KW oder WW auf) nichts tun, da sind wir in der Mitte der Hysterese
				
				temp = KW_VENTIL & WW_VENTIL;
				
				if (temp == VENTIL_ZU)
				{
					WW_VENTIL = VENTIL_AUF;
					overl_ak_wk = 0;	// overlap warm->kalt freigeben
				}
			}
			
			// bei Vorwsche usprngliche Temperatur wiederherstellen
			if (VORWAESCHE)
			{
				ZielTemp = Solltemp;	
			}
			
		}
		else if (temp2) // bei Splstatus/Fehler nur Kaltwasser
		{
			KW_VENTIL = VENTIL_AUF;	
			NOP;				// zwei Nops wegen Timing read-mod-write am port
			NOP;
			WW_VENTIL = VENTIL_ZU;
			
			v_overl_cnt = 0; 	// alle berlappvorgnge aus
			
			// wenn die Zulauftemperatur KW aktueller (kleiner) ist, ruhig mal updaten
			if (ZulaufTemp.low8 < KWTemp)
			{
				KWTemp = ZulaufTemp.low8;
			}
		}	
	}
	else // Einlauf zu oder Fllzeit berschritten -> unsere Ventile auch zu
	{
		KW_VENTIL = VENTIL_ZU;
		NOP;				// zwei Nops wegen Timing read-mod-write am port
		NOP;
		WW_VENTIL = VENTIL_ZU;
		
		v_overl_cnt = 0; 	// alle berlappvorgnge aus
	}
	
	
#pragma updateBank 1 /* ON, all cores */
}
	
//
// --- Display update ---
//
// LCD-Ausgaben
// liest: Update_Prog, ZielTemp, VORWAESCHE, KW_VENTIL, WW_VENTIL, akt_status, V_IN1, TrommelTemp
// schreibt: KWTemp, WWTemp, Update_Prog
//
void update_LCD(void)
{
#pragma updateBank 0 /* OFF, all cores */

	// Hat sich was am Programm getan -> Update
	if (Update_Prog)
	{
		LCDPos(0,6);
		// Zieltemperatur updaten
		if (ZielTemp == 0)
		{
			LCDString("kalt");
		}
		else
		{
			bin2dec999(ZielTemp);
			Text[0]=' ';			
			Text[3]=' ';	// Rest von KALT lschen
			Text[4]=0;
			LCDString(Text);
		}
		
		// Vorwsche updaten
		LCDPos(0,11);
		if (VORWAESCHE)
		{
			LCDString("Vorw");
		}
		else
		{
			LCDString("    ");
		}
		
		Update_Prog = 0; // Update besttigt
	}
	
	// -- dann Update zweite Zeile --
	
	
	// Kaltwasserteil-Update
	bin2dec999(KWTemp);
	if (KW_VENTIL == VENTIL_AUF)
	{
		Text[3]='+';	
	}
	else
	{
		Text[3]='-';
	}
	Text[0]='K';
	Text[4] = ' ';
	Text[5] = 0;
	
	LCDPos(1,0);
	LCDString(Text);
	
	// Warmwasserteil-Update
	bin2dec999(WWTemp);
	if (WW_VENTIL == VENTIL_AUF)
	{
		Text[3]='+';	
	}
	else
	{
		Text[3]='-';
	}
	Text[0]='W';
	Text[4] = ' ';
	Text[5] = 0;
	
	LCDPos(1,5);
	LCDString(Text);
	
	if (V_IN1)
	{
		Text[0]='Z';	
	}
	else
	{
		Text[0]='-';
		
	}
	Text[1] = 0;
	LCDString(Text);
	
	//Trommeltemperatur nur beim Mischen angeben (sonst unbekannt)
	if (akt_status == ST_MIX)
	{
		bin2dec999(TrommelTemp);
		Text[0]=' ';
		Text[3]=0;
		LCDString(Text);
	}
	else if (akt_status == ST_PURGE)
	{
		LCDString("   ");
	}
	else if (akt_status == ST_FAIL) // Fehlerstatus ordentlich ausgeben
	{
		LCDPos(0,0);
		
		LCDString("Err: ");
		
		switch (fcode)
		{
			case F_OVERT:	LCDString("Uebertemp");
					break;
			case F_LOWT:	LCDString("Untertemp");
					break;
			case F_FUELLZEIT:	LCDString("Fuellzeitmax");
					break;
			default:	LCDString("unbekannt");
					break;
		}
	}
	
	
	
	/* Ventil 2 erstmal nicht benutzt 
	if (V_IN2)
	{
		Text[9]='Z';	
	}
	else
	{
		Text[9]='-';
	}
	Text[10]=0; */
	
	// Wenn Backlight gefordert ist, anmachen
	if (lcdlight)
	{
		LCD_BACKLIT = 1;
		lcdlight--;
	}
	else
	{
		LCD_BACKLIT = 0;	
	}
	
	
#pragma updateBank 1 /* ON, all cores */
}

//
// --- Zeitcheck ---
//
// Gesamtzeit begrenzen
// liest: Gesamtzeit
// schreibt: -
//
void Zeitcheck(void)
{
	// 16k s = 4,5 Stunden... Laufzeit zuende-> alles aus, Strom sparen.
	if (Gesamtzeit.14)
	{
		LCDCls();
		LCDPos(1,0);
		LCDString(" Beendet");
		
		LCD_BACKLIT = 0; // Licht aus
		NOP;		// zwei Nops wegen Timing read-mod-write am port
		NOP;
		LED_OFF = 1;	 // LED auf jeden Fall aus
		
		KW_VENTIL = VENTIL_ZU;
		NOP;		// zwei Nops wegen Timing read-mod-write am port
		NOP;
		WW_VENTIL = VENTIL_ZU;
		
		
		ADON = 0; 	// ADC ausschalten
		sleep();	// Strom sparen...
		
		while (1);
	}
}


/*------------------------------------------------------------------*/
/*     M     M    A    II  N   N                                    */
/*     MM   MM   A A   II  NN  N                                    */
/*     M M M M  A   A  II  N N N                                    */
/*     M  M  M  AAAAA  II  N  NN                                    */
/*     M     M  A   A  II  N   N                                    */
/*------------------------------------------------------------------*/

void main (void)
{
		
	init();
	
	LCDInit();	  // LCD Initialisieren
		
	LCDCls();	 // LCD lschen
	
	LCDPos(0,0);
	LCDString("Temp:");
	
	LCDPos(1,0);
	LCDString("WIMix 1.2");
	    
	temp_messen(); // Temperatur messen 

	KWTemp = ZulaufTemp.low8; // KW und WW erstmals belegen. Auflsung ZulaufTemp: etwa 0.5
	WWTemp = KWTemp;
	
	LED_OFF = 1; // LED aus
		
	while (1)
	{	
		check_input();	// Ventil- und Tastenstatus lesen
		
		Delay1ms(2);	// Zeitscheiben-Einteilung
		
		temp_messen();	// Temperatur messen
		
		update_tasten(); // Tasten einlesen
		
		update_state();	// Status testen
		
		update_vent();	// Ventilstatus bestimmen und Ventile steuern
		
		if (changed)
		{
			update_LCD();	// Display updaten
			LED_OFF ^= 1;
			Zeitcheck();	// Ende der Laufzeit testen -> Endlos-Schleife
		}
		
	}; // von while(immer)


}
